#global imports
from db.entities.layer_adapter import LayerAdapter
from db.repository import Repository
from db.schema_repository import SchemaRepository
from db.use_case_repository import UseCaseRepository
from services.layer_adapter_service import LayerAdapterService

import json
from flask import Response, request

schema_repository = SchemaRepository()
layer_repository = Repository()
use_case_repository = UseCaseRepository()

def all():
    return [layer.to_serializable_dict() for layer in layer_repository.all()]

def get_all_for_use_case(use_case: str):
    '''
    get all layers assigned to the given use_case
    '''
    use_case_repository.put(use_case)
    return [layer.to_serializable_dict() for layer in layer_repository.all_for_use_case(use_case)]

def add_complete():
    '''
    add a layer already containing attribute mappings to the BD.
    '''
    data = request.json

    if "name" not in data or "properties" not in data or "cluster_properties" not in data or "use_case" not in data:
        return Response(status=400, response=f"Field missing! Fields required: (name, properties, cluster_properties)")

    use_case_repository.put(data["use_case"])
    layer = layer_repository.one(data["name"], data["use_case"])

    if layer != None:
        return Response(status=400, response=f'Layer with name "{data["name"]}" already exists!')

    # check if schema contains mapping
    try:
        layer_new = LayerAdapter.from_serializable_dict(data)
    except BaseException as e:
        print(f"Exception: {e}")
        return Response(status=400)
    
    try:
        LayerAdapterService.add_complete(layer_new)
    except ValueError as e:
        return Response(status = 400, response=f"{e}")
    
    return Response(status=200)

def one(name: str, use_case: str):
    '''
    fetch a single layer from the DB

    @params:
        name        - Required : unique identifier for the layer
        use_case    - Required : String-identifier for the use-case
    '''
    use_case_repository.put(use_case)
    layer = layer_repository.one(name, use_case)

    if layer == None:
        return Response(status=404, response=f"Layer with name '{name}' does not exist!")

    return Response(status=200, response=json.dumps(layer.to_serializable_dict()))

def add_empty(name: str, use_case: str):
    '''
    add an empty layer to the DB

    @params:
        name        - Required : unique identifier for the layer
        use_case    - Required : String-identifier for the use-case
    '''
    use_case_repository.put(use_case)

    if layer_repository.one(name, use_case) != None:
        return Response(status=400, response=f"Layer with name '{name}' already exists!")

    layer_repository.add_empty(name, use_case)
    return Response(status=200)

def add_cluster_mapping(name: str, use_case: str):
    '''
    add a mapped property to the list of properties to cluster with

    @params:
        name        - Required : unique identifier for the layer
        use_case    - Required : String-identifier for the use-case
    '''
    use_case_repository.put(use_case)

    layer = layer_repository.one(name, use_case)

    if layer == None:
        return Response(status=404, response=f"Layer with name '{name}' does not exist!")

    data = request.json

    if "attribute" not in data:
        return Response(status=400, response=f"Field missing! Fields required: (attribute)")

    try:
        layer.add_cluster_mapping(data["attribute"])
        layer_repository.update(layer)
        return Response(status=200)
    except:
        return Response(status=400, response=f'{data["attribute"]} is no attribute in the layer!')

def delete_cluster_mapping(name: str, use_case: str):
    '''
    remove a mapped property from the list of properties to cluster with

    @params:
        name        - Required : unique identifier for the layer
        use_case    - Required : String-identifier for the use-case
    '''
    use_case_repository.put(use_case)

    layer = layer_repository.one(name, use_case)

    if layer == None:
        return Response(status=404, response=f"Layer with name '{name}' does not exist!")

    data = request.json

    if "attribute" not in data:
        return Response(status=400, response=f"Field missing! Fields required: (attribute)")

    try:
        layer.delete_cluster_mapping(data["attribute"])
        layer_repository.update(layer)
        return Response(status=200)
    except ValueError as e:
        print(e)
        return Response(status=400, response=f'{data["attribute"]} is no attribute in the layer!')

def add_mapping_for_all(use_case: str):
    '''
    add a new mapping to each layer of the use-case

    @params:
        use_case    - Required : String-identifier for the use-case
    '''
    use_case_repository.put(use_case)

    data = request.json

    if "internal" not in data:
        return Response(status=400, response=f"Field missing! Fields required: (internal)")
    
    # check if schema contains mapping
    schema = schema_repository.put(use_case)

    if data["internal"] not in schema.mappings:
        return Response(status=400, response=f'{data["internal"]} is not existent in the schema!')

    layers = layer_repository.all_for_use_case(use_case)

    for layer in layers:
        layer.add_mapping(data["internal"])

        try:
            LayerAdapterService.check_layer(layer)
        except ValueError as e:
            return Response(status = 400, response=f"{e}")

        layer_repository.update(layer)

    return Response(status=200)

def delete_mapping_for_all(use_case: str):
    '''
    delete a mapping from each layer in the use-case

    @params:
        use_case    - Required : String-identifier for the use-case
    '''
    use_case_repository.put(use_case)

    data = request.json

    if "internal" not in data:
        return Response(status=400, response=f"Field missing! Fields required: (internal)")
    
    layers = layer_repository.all_for_use_case(use_case)

    for layer in layers:
        try:
            layer.delete_mapping(data["internal"])
        except ValueError:
            continue
        layer_repository.update(layer)

    return Response(status=200)

def delete_mapping(name: str, use_case: str):
    '''
    delete a mapping from the layer identified by the internal representation

    @params:
        name        - Required : unique identifier for the layer
        use_case    - Required : String-identifier for the use-case
    '''
    use_case_repository.put(use_case)

    layer = layer_repository.one(name, use_case)

    if layer == None:
        return Response(status=404, response=f"Layer with name '{name}' does not exist!")

    data = request.json

    if "internal" not in data:
        return Response(status=400, response=f"Field missing! Fields required: (internal)")

    try:
        layer.delete_mapping(data["internal"])
        layer_repository.update(layer)
    except ValueError:
        return Response(status=400, response=f'{data["internal"]} is not a property of the layer!')

    return Response(status=200)

def add_mapping(name: str, use_case: str):
    '''
    add a new mapping to the layer identified by name

    @params:
        name        - Required : unique identifier for the layer
        use_case    - Required : String-identifier for the use-case
    '''
    use_case_repository.put(use_case)

    layer = layer_repository.one(name, use_case)

    if layer == None:
        return Response(status=404, response=f"Layer with name '{name}' does not exist!")

    data = request.json

    if "internal" not in data:
        return Response(status=400, response=f"Field missing! Fields required: (internal)")

    # check if schema contains mapping
    schema = schema_repository.put(use_case)
    if data["internal"] not in schema.mappings:
        return Response(status=400, response=f'{data["internal"]} is not existent in the schema!')

    layer.add_mapping(data["internal"])
    layer_repository.update(layer)

    return Response(status=200)

def delete_one(name: str, use_case: str):
    '''
    delete a layer and all its mappings from the Db

    @params:
        name        - Required : unique identifier for the layer
        use_case    - Required : String-identifier for the use-case
    '''
    use_case_repository.put(use_case)
    
    layer = layer_repository.one(name, use_case)

    if layer == None:
        return Response(status=404, response=f"Layer with name '{name}' does not exist!")

    layer_repository.delete_all_with_name_and_use_case(layer)
    return Response(status=200)

def delete_all_layers():
    '''
    delete all layers from the DB
    '''
    layer_repository.delete_all()
    return Response(status=200)